
# parsetab.py
# This file is automatically generated. Do not edit.
# pylint: disable=W,C,R
_tabversion = '3.10'

_lr_method = 'LALR'

_lr_signature = 'leftORleftANDrightNOTleftGREATERLESSGREATER_EQUALLESS_EQUALEQUALNOT_EQUALREGEX_MATCHAND EQUAL GREATER GREATER_EQUAL LESS LESS_EQUAL LPAREN NOT NOT_EQUAL NUMBER OR REGEX_MATCH RPAREN STRING TOPICexpr : termexpr : expr AND exprexpr : expr OR exprexpr : NOT exprexpr : LPAREN expr RPARENterm : TOPIC LESS NUMBERterm : TOPIC LESS_EQUAL NUMBERterm : TOPIC GREATER_EQUAL NUMBERterm : TOPIC GREATER NUMBERterm : TOPIC EQUAL literalterm : TOPIC NOT_EQUAL literalterm : TOPIC REGEX_MATCH STRINGliteral : NUMBER\n                   | STRING'
    
_lr_action_items = {'NOT':([0,3,4,6,7,],[3,3,3,3,3,]),'LPAREN':([0,3,4,6,7,],[4,4,4,4,4,]),'TOPIC':([0,3,4,6,7,],[5,5,5,5,5,]),'$end':([1,2,8,17,18,19,20,21,22,23,24,25,26,27,28,],[0,-1,-4,-2,-3,-5,-6,-7,-8,-9,-10,-13,-14,-11,-12,]),'AND':([1,2,8,9,17,18,19,20,21,22,23,24,25,26,27,28,],[6,-1,-4,6,-2,6,-5,-6,-7,-8,-9,-10,-13,-14,-11,-12,]),'OR':([1,2,8,9,17,18,19,20,21,22,23,24,25,26,27,28,],[7,-1,-4,7,-2,-3,-5,-6,-7,-8,-9,-10,-13,-14,-11,-12,]),'RPAREN':([2,8,9,17,18,19,20,21,22,23,24,25,26,27,28,],[-1,-4,19,-2,-3,-5,-6,-7,-8,-9,-10,-13,-14,-11,-12,]),'LESS':([5,],[10,]),'LESS_EQUAL':([5,],[11,]),'GREATER_EQUAL':([5,],[12,]),'GREATER':([5,],[13,]),'EQUAL':([5,],[14,]),'NOT_EQUAL':([5,],[15,]),'REGEX_MATCH':([5,],[16,]),'NUMBER':([10,11,12,13,14,15,],[20,21,22,23,25,25,]),'STRING':([14,15,16,],[26,26,28,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'expr':([0,3,4,6,7,],[1,8,9,17,18,]),'term':([0,3,4,6,7,],[2,2,2,2,2,]),'literal':([14,15,],[24,27,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> expr","S'",1,None,None,None),
  ('expr -> term','expr',1,'p_expr_term','expr.py',87),
  ('expr -> expr AND expr','expr',3,'p_expr_and','expr.py',91),
  ('expr -> expr OR expr','expr',3,'p_expr_or','expr.py',96),
  ('expr -> NOT expr','expr',2,'p_expr_not','expr.py',101),
  ('expr -> LPAREN expr RPAREN','expr',3,'p_expr_parenthesis','expr.py',106),
  ('term -> TOPIC LESS NUMBER','term',3,'p_term_topic_less_number','expr.py',110),
  ('term -> TOPIC LESS_EQUAL NUMBER','term',3,'p_term_topic_less_equal_number','expr.py',115),
  ('term -> TOPIC GREATER_EQUAL NUMBER','term',3,'p_term_topic_greater_equal_number','expr.py',120),
  ('term -> TOPIC GREATER NUMBER','term',3,'p_term_topic_greater_number','expr.py',125),
  ('term -> TOPIC EQUAL literal','term',3,'p_term_topic_equal_literal','expr.py',130),
  ('term -> TOPIC NOT_EQUAL literal','term',3,'p_term_topic_not_equal_literal','expr.py',136),
  ('term -> TOPIC REGEX_MATCH STRING','term',3,'p_term_topic_regex_match_string','expr.py',142),
  ('literal -> NUMBER','literal',1,'p_literal','expr.py',148),
  ('literal -> STRING','literal',1,'p_literal','expr.py',149),
]



from functools import partial, wraps
import logging
from pprint import pformat
from signal import signal, SIGUSR1
from . expr import Parser
from . mqtt import topic


__all__ = ('StateAwareMixin', 'conditional', 'when')


class StateAwareMixin:
    ''' Mixin for stateful MQTT clients.
        Status updates are recorded in-memory from MQTT topics,
        e.g. `status/#`.
        The message payload for status updates is JSON-converted if possible.
        The last known state is available in `self.current_state`.
        Subclasses may define handler methods using the @when decorator`.
    '''
    
    conditions = {}
    expr_parser = Parser()
    update_log_level = logging.INFO
    
    
    def __init__(self, **kw):
        'Register topics and the state callcack.'
        
        super().__init__(**kw)
        self.current_state = {}

        status_topic = self.get_config().get('status_topic')
        assert status_topic, 'status_topic not found in configuration'

        # Subscribe to status updates
        register = topic(status_topic, payload_converter=self.decode_json)
        register(self.update_status)

        # Dump curent state on USR1 signal
        signal(SIGUSR1, self.dump_state)


    @staticmethod
    def update_status(self, _userdata, msg):
        ''' Track the global state,
            and invoke handler methods defined by subclasses
            with the message payload.
        '''
        if self.on_status_update(msg.topic, msg.payload):
            self.invoke_handlers(msg.topic, msg.payload)


    def on_status_update(self, topic, payload):
        ''' Keep the global state in-memory.
            Returns a path to the updated attribute in `self.current_state`
            when the state has changed, or `None` otherwise.
        '''
        # Update only if the value has changed
        if self.current_state.get(topic) != payload:
            self.current_state[topic] = payload
            self.log.log(self.update_log_level, 'Updated: %s = %s', topic, payload)
            return topic
    
    
    def invoke_handlers(self, topic, payload):
        'Run through conditions and invoke appropriate handlers'
        for expr, method in self.conditions.items():
            if topic in expr.keys and expr.keys <= self.current_state.keys():
                method(self)


    def publish(self, topic, payload=None, **kw):
        'Avoid redundant updates'
        old_state = self.current_state.get(topic)
        if old_state is not None:
            if old_state == payload: return

            cast = type(old_state)
            try:
                if cast(payload) == old_state: return
            except: pass
        return super().publish(topic, payload, **kw)


    def dump_state(self, _signal, _frame):
        'Print status information'
        print('Current state: ' + pformat(self.current_state))


def conditional(expression):
    ''' Decorated a status handler method which is invoked
        with the value of the expression
        whenever the expression depends on an updated topic.
        For the expression grammar, see the docstrings in `expr.py`.
    '''
    
    predicate = StateAwareMixin.expr_parser.parse(expression)
    
    def wrapper(method):
        id = method.func.__name__ if type(method) is partial else method.__name__
        
        @wraps(method)
        def wrapped(self):
            last_state = predicate.last_state
            condition = predicate(self.current_state)
            if condition != last_state:
                self.log.info('Invoking: %s(%s)', id, condition)
                method(self, condition)
        StateAwareMixin.conditions[predicate] = wrapped
        return method
    return wrapper


def when(expression):
    ''' Decorate a status handler method which is invoked
        when the expression depends on an updated topic
        and when it evaluates to `True`.
        For the expression grammar, see the docstrings in `expr.py`.
    '''
    
    predicate = StateAwareMixin.expr_parser.parse(expression)
    
    def wrapper(method):
        id = method.func.__name__ if type(method) is partial else method.__name__
        
        @wraps(method)
        def wrapped(self):
            last_state = predicate.last_state
            condition = predicate(self.current_state)
            if condition != last_state:
                if condition:
                    self.log.info('Invoking: %s', id)
                    method(self)
                else:
                    self.log.debug('Skipping: %s', id)
        StateAwareMixin.conditions[predicate] = wrapped
        return method
    return wrapper



from collections import namedtuple
from croniter import croniter
from datetime import datetime
from functools import partial, wraps
from heapq import heappush, heappop, heapify
from localtime import now
from signal import signal, SIGUSR2
import logging, threading, os, time


__all__ = ('cron', 'Scheduler')


class Tasks:
    'A synchronised priority queue of scheduled tasks'

    class Task(namedtuple('Task', 'when id func')): # sorted by when
        def __repr__(self):
            return 'Task(%s %s)' % (self.when, self.id)


    def __init__(self, resolution=0.1, daemon=False, log_level=logging.INFO):
        self.log = logging.getLogger(self.__class__.__name__)
        self.log.setLevel(log_level)
        self.resolution = resolution
        self.tasks = []
        self.mutex = threading.Lock()
        self.worker = threading.Thread(target=self.run, daemon=daemon, name='task_list')
        self.worker.start()

        
    def add(self, task):
        'Add a task to the queue'
        self.cancel(task.id)
        with self.mutex:
            heappush(self.tasks, task)
            self.log.debug('Added task (%d/%d): %s',
                self.tasks.index(task) + 1,
                len(self.tasks),
                task)

    
    def create(self, func, when=None, id=None):
        'Add a task from parameters'
        self.add(self.Task(when or now(), id or func.__name__, func))


    def addAll(self, tasks):
        'Add all tasks from an iterable'
        for task in tasks: self.add(task)
            

    def __len__(self):
        'Number of tasks in the queue'
        return len(self.tasks)
        

    def __contains__(self, element):
        'Membership test'
        if type(element) is str:
            return any(task.id == element for task in self.tasks)
        return element in self.tasks
        
    
    def __repr__(self):
        return '--- %d Tasks:\n%s' % (len(self),
            '\n'.join('%2d: %s' % (n + 1, t)
                for n, t in enumerate(sorted(self.tasks,
                    key=lambda t: (t.when, t.id)))))


    def cancel(self, id):
        'Remove all tasks with the given ID'
        if id not in self: return
        with self.mutex:
            for pos in reversed(list(p for p,t in enumerate(self.tasks) if t.id == id)):
                task = self.tasks[pos]
                del self.tasks[pos]
                self.log.debug('Removed task: (%d/%d) %s',
                    pos + 1, len(self.tasks) + 1, task)
            heapify(self.tasks)
        

    def next(self):
        'Pop the next available task from the queue'
        if not self.tasks: return
        with self.mutex:
            if self.tasks[0].when <= now():
                return heappop(self.tasks)


    def run(self):
        'Execute pending tasks'
        self.log.debug('Starting tasks...')
        self.running = True

        while self.running:
            task = self.next()
            if task: self.execute(task)
            else: time.sleep(self.resolution)

        self.log.debug('Stopped tasks')


    def execute(self, task):        
        def wrapper():
            try:
                start = now()
                task.func()
                duration = (now() - start).total_seconds()
                self.log.debug('Task %s took %.2fs', task.id, duration)
            except:
                self.log.exception('Error in task: %s', task.id)
                
        self.log.debug('Executing task: %s', task.id)
        threading.Thread(target=wrapper, name=task.id, daemon=True).start()
        

    def stop(self):
        'Stop polling, must be called from the main thread.'
        if self.running:
            self.log.debug('Stopping tasks (%d pending)' % len(self))
            self.running = False
            self.worker.join()


class Scheduler:
    'Mixin class for time-based task execution'
    
    startup_tasks = set()
    
    def __init__(self):
        super().__init__()
        self.log.debug('--- Starting with PID: %d ---', os.getpid())
        signal(SIGUSR2, self.dump_tasks)

        self.tasks = Tasks(
            resolution=1,
            daemon=True,
            log_level=self.log.level)

        for when, method in self.startup_tasks:
            self.tasks.create(partial(method, self), when, method.__name__)


    def run(self):
        'Clean up on exit'
        try: super().run()
        finally: self.tasks.stop()
        
    
    def dump_tasks(self, _signal, _frame):
        'List pending tasks'
        print('Pending tasks: %s' % self.tasks)
    

def cron(schedule, log_level=logging.DEBUG):
    'Decorator for periodic tasks'
    sequence = croniter(schedule, now())
    
    def wrapper(method):
        @wraps(method)
        def periodic_task(self):
            if method.__name__ not in self.tasks:
                when = sequence.get_next(datetime)
                while when < now():
                    when = sequence.get_next(datetime)
                self.tasks.create(partial(periodic_task, self),
                    when, method.__name__)
            self.log.log(log_level, 'Invoking: %s' % method.__name__)
            method(self)
            
        Scheduler.startup_tasks.add((sequence.get_next(datetime), periodic_task))
        return periodic_task
    return wrapper



from . intent import IntentPayload
from . logging import LoggingMixin
from . snips import *
from basecmd import BaseCmd
import logging


if __name__ == '__main__': # demo code

    class Logger(BaseCmd, LoggingMixin, SnipsClient):
            
        @on_intent('#')
        # Do not use @intent here because it ends the session,
        # and thus interferes with other intent handlers
        def intent_logger(self, userdata, msg):
            self.log_intent(IntentPayload(msg.payload), level=logging.INFO)

        @on_end_session()
        @on_continue_session()
        def response_logger(self, userdata, msg):
            self.log_response(msg.payload.get('text'), level=logging.INFO)

    Logger().run()



from colors import cyan, green, magenta, red, yellow
import logging


__all__ = ('LoggingMixin', )


class LoggingMixin:
    'Logging for Snips events'
    
    INDENT = 10
    tty_log = None

    def colored_log(self, level, format, *args, color=None):
        if self.tty_log is None:
            self.tty_log = logging.getLogger().handlers[0].stream.isatty()
        if color and self.tty_log: args = map(color, args)
        self.log.log(level, format, *args)
    

    def tabular_log(self, level, key, value, color=None, width=INDENT):
        label = '%-*s' % (width, key)
        self.colored_log(level, '%s %s', label, str(value), color=color)
    
    
    def log_intent(self, payload, level=logging.DEBUG):
        'Log an intent message'
        self.tabular_log(level, 'intent', '%s, confidence: %.1f' % (
            red(payload.intent.intent_name, style='bold'),
            payload.intent.confidence_score), color=green)
        for k in ('site_id', 'input'):
            self.tabular_log(level, k, getattr(payload, k), color=cyan)
        for name, slot in payload.slots.items():
            self.tabular_log(level, name, slot.value, color=magenta)
        if payload.custom_data:
            self.tabular_log(level, 'data', payload.custom_data, color=yellow)
            
    
    def log_response(self, response, level=logging.DEBUG):
        'Log an action response'
        if response: self.tabular_log(level, 'answer',
            red(response, style='bold'), color=green)



from collections import namedtuple
import ply.lex as lex
import ply.yacc as yacc
import re


class Parser:
    'Parse very simple expressions into callable boolean conditions'
    
    class Expr(namedtuple('Expr', 'keys expr')):
        '''
            Container for a callable boolean predicate
            along with the topic names being evaluated
        '''
        last_state = None
        
        def __call__(self, *args, **kw):
            'Syntactic sugar'
            self.last_state = self.expr(*args, **kw)
            return self.last_state
            
        def __hash__(self):
            return hash(self.expr)


    tokens = (
        'TOPIC',
        'LESS',
        'LESS_EQUAL',
        'EQUAL',
        'REGEX_MATCH',
        'GREATER_EQUAL',
        'GREATER',
        'NOT_EQUAL',
        'NUMBER',
        'NOT',
        'AND',
        'OR',
        'LPAREN',
        'RPAREN',
        'STRING',
    )

    t_TOPIC         = r'[\w_-]+(/[\w_-]+)+'
    t_GREATER       = r'>'
    t_LESS          = r'<'
    t_EQUAL         = r'=='
    t_REGEX_MATCH   = r'~='
    t_GREATER_EQUAL = r'>='
    t_LESS_EQUAL    = r'<='
    t_NOT_EQUAL     = r'!='
    t_NOT           = r'not'
    t_AND           = r'and'
    t_OR            = r'or'
    t_LPAREN        = r'\('
    t_RPAREN        = r'\)'

    # Ignored characters
    t_ignore        = ' \t\r\n'

    def t_NUMBER(self, t):
        r'(\+|-)?\d+(\.\d+)?'
        t.value = float(t.value)    
        return t

    def t_STRING(self, t):
        r"'[^']*'"
        t.value = t.value[1:-1]
        return t
        
    # Track line numbers
    def t_newline(self, t):
        r'\n+'
        t.lexer.lineno += len(t.value)

    def t_error(self, t):
        assert False, "Illegal character '%s'" % t.value[0]

    #-#-# Grammar rules below #-#-#

    precedence = (
        ('left',  'OR'),
        ('left',  'AND'),
        ('right', 'NOT'),
        ('left',  'GREATER', 'LESS', 'GREATER_EQUAL', 'LESS_EQUAL',
                  'EQUAL', 'NOT_EQUAL', 'REGEX_MATCH'),
    )
    
    def p_expr_term(self, p):
        'expr : term'
        p[0] = p[1]

    def p_expr_and(self, p):
        'expr : expr AND expr'
        lhs, rhs = p[1], p[3]
        p[0] = self.Expr(lhs.keys | rhs.keys, lambda s: lhs(s) and rhs(s))

    def p_expr_or(self, p):
        'expr : expr OR expr'
        lhs, rhs = p[1], p[3]
        p[0] = self.Expr(lhs.keys | rhs.keys, lambda s: lhs(s) or rhs(s))

    def p_expr_not(self, p):
        'expr : NOT expr'
        cond = p[2]
        p[0] = self.Expr(cond.keys, lambda s: not cond(s))

    def p_expr_parenthesis(self, p):
        'expr : LPAREN expr RPAREN'
        p[0] = p[2]

    def p_term_topic_less_number(self, p):
        'term : TOPIC LESS NUMBER'
        lhs, rhs = p[1], p[3]
        p[0] = self.Expr(set([lhs]), lambda s: float(s[lhs]) < rhs)
        
    def p_term_topic_less_equal_number(self, p):
        'term : TOPIC LESS_EQUAL NUMBER'
        lhs, rhs = p[1], p[3]
        p[0] = self.Expr(set([lhs]), lambda s: float(s[lhs]) <= rhs)
        
    def p_term_topic_greater_equal_number(self, p):
        'term : TOPIC GREATER_EQUAL NUMBER'
        lhs, rhs = p[1], p[3]
        p[0] = self.Expr(set([lhs]), lambda s: float(s[lhs]) >= rhs)
        
    def p_term_topic_greater_number(self, p):
        'term : TOPIC GREATER NUMBER'
        lhs, rhs = p[1], p[3]
        p[0] = self.Expr(set([lhs]), lambda s: float(s[lhs]) > rhs)
        
    def p_term_topic_equal_literal(self, p):
        'term : TOPIC EQUAL literal'
        lhs, rhs = p[1], p[3]
        cast = type(rhs)
        p[0] = self.Expr(set([lhs]), lambda s: cast(s[lhs]) == rhs)
        
    def p_term_topic_not_equal_literal(self, p):
        'term : TOPIC NOT_EQUAL literal'
        lhs, rhs = p[1], p[3]
        cast = type(rhs)
        p[0] = self.Expr(set([lhs]), lambda s: cast(s[lhs]) != rhs)
        
    def p_term_topic_regex_match_string(self, p):
        'term : TOPIC REGEX_MATCH STRING'
        lhs, rhs = p[1], p[3]
        pattern = re.compile(rhs)
        p[0] = self.Expr(set([lhs]), lambda s: pattern.search(str(s[lhs])) is not None)
        
    def p_literal(self, p):
        '''literal : NUMBER
                   | STRING'''
        p[0] = p[1]

    def p_error(self, p):
        assert False, 'Syntax error at: %s' % (p or 'EOF')

    def __init__(self, **kwargs):
        self.lexer = lex.lex(module=self)
        self.parser = yacc.yacc(module=self, **kwargs)
        
    def parse(self, text, **kwargs):
        return self.parser.parse(text, lexer=self.lexer, **kwargs)



from expr import Parser
import unittest
    
class ExprTest(unittest.TestCase):
    
    parser = Parser()

    def test_less(self):
        expr = self.parser.parse('topic/a < 1')
        self.assertEqual(expr.keys, set(['topic/a']))
        self.assertFalse(expr({ 'topic/a' : 1 }))
        self.assertFalse(expr({ 'topic/a' : '1' }))
        self.assertTrue(expr({ 'topic/a' : 0 }))
        self.assertTrue(expr({ 'topic/a' : '0' }))
        self.assertTrue(expr.last_state)
        self.assertRaises(KeyError, expr, {})
        self.assertRaises(ValueError, expr, { 'topic/a' : 'foo' })

    def test_less_equal(self):
        expr = self.parser.parse('topic/a <= 1')
        self.assertEqual(expr.keys, set(['topic/a']))
        self.assertFalse(expr({ 'topic/a' : 2 }))
        self.assertFalse(expr({ 'topic/a' : '2' }))
        self.assertTrue(expr({ 'topic/a' : 1 }))
        self.assertTrue(expr({ 'topic/a' : '1' }))
        self.assertTrue(expr.last_state)
        self.assertRaises(KeyError, expr, {})
        self.assertRaises(ValueError, expr, { 'topic/a' : 'foo' })

    def test_greater_equal(self):
        expr = self.parser.parse('topic/a >= 1')
        self.assertEqual(expr.keys, set(['topic/a']))
        self.assertFalse(expr({ 'topic/a' : 0 }))
        self.assertFalse(expr({ 'topic/a' : '0' }))
        self.assertTrue(expr({ 'topic/a' : 1 }))
        self.assertTrue(expr({ 'topic/a' : '1' }))
        self.assertTrue(expr.last_state)
        self.assertRaises(KeyError, expr, {})
        self.assertRaises(ValueError, expr, { 'topic/a' : 'foo' })

    def test_greater(self):
        expr = self.parser.parse('topic/a > 1')
        self.assertEqual(expr.keys, set(['topic/a']))
        self.assertFalse(expr({ 'topic/a' : 1 }))
        self.assertFalse(expr({ 'topic/a' : '1' }))
        self.assertTrue(expr({ 'topic/a' : 2 }))
        self.assertTrue(expr({ 'topic/a' : '2' }))
        self.assertTrue(expr.last_state)
        self.assertRaises(KeyError, expr, {})
        self.assertRaises(ValueError, expr, { 'topic/a' : 'foo' })

    def test_negative_greater(self):
        expr = self.parser.parse('topic/a > -1')
        self.assertEqual(expr.keys, set(['topic/a']))
        self.assertFalse(expr({ 'topic/a' : -1 }))
        self.assertFalse(expr({ 'topic/a' : '-1' }))
        self.assertTrue(expr({ 'topic/a' : 0 }))
        self.assertTrue(expr({ 'topic/a' : '0' }))
        self.assertTrue(expr.last_state)
        self.assertRaises(KeyError, expr, {})
        self.assertRaises(ValueError, expr, { 'topic/a' : 'foo' })

    def test_equal_number(self):
        expr = self.parser.parse('topic/a == 1')
        self.assertEqual(expr.keys, set(['topic/a']))
        self.assertFalse(expr({ 'topic/a' : 0 }))
        self.assertFalse(expr({ 'topic/a' : '0' }))
        self.assertTrue(expr({ 'topic/a' : 1 }))
        self.assertTrue(expr({ 'topic/a' : '1' }))
        self.assertTrue(expr.last_state)
        self.assertRaises(KeyError, expr, {})
        self.assertRaises(ValueError, expr, { 'topic/a' : 'foo' })

    def test_equal_string(self):
        expr = self.parser.parse("topic/a == '1'")
        self.assertEqual(expr.keys, set(['topic/a']))
        self.assertFalse(expr({ 'topic/a' : '' }))
        self.assertTrue(expr({ 'topic/a' : '1' }))
        self.assertTrue(expr({ 'topic/a' : 1 }))
        self.assertTrue(expr.last_state)
        self.assertRaises(KeyError, expr, {})

    def test_regex_match(self):
        expr = self.parser.parse("topic/a ~= 'o.?o'")
        self.assertEqual(expr.keys, set(['topic/a']))
        self.assertTrue(expr({ 'topic/a' : 'zoo' }))
        self.assertTrue(expr({ 'topic/a' : 'bozo' }))
        self.assertFalse(expr({ 'topic/a' : 'bar' }))
        self.assertFalse(expr.last_state)
        self.assertRaises(KeyError, expr, {})

    def test_not_equal_number(self):
        expr = self.parser.parse('topic/a != 1')
        self.assertEqual(expr.keys, set(['topic/a']))
        self.assertFalse(expr({ 'topic/a' : 1 }))
        self.assertFalse(expr({ 'topic/a' : '1' }))
        self.assertTrue(expr({ 'topic/a' : 0 }))
        self.assertTrue(expr({ 'topic/a' : '0' }))
        self.assertTrue(expr.last_state)
        self.assertRaises(KeyError, expr, {})
        self.assertRaises(ValueError, expr, { 'topic/a' : 'foo' })

    def test_not_equal_string(self):
        expr = self.parser.parse("topic/a != '1'")
        self.assertEqual(expr.keys, set(['topic/a']))
        self.assertFalse(expr({ 'topic/a' : '1' }))
        self.assertTrue(expr({ 'topic/a' : '' }))
        self.assertTrue(expr({ 'topic/a' : 0 }))
        self.assertTrue(expr.last_state)
        self.assertRaises(KeyError, expr, {})

    def test_not(self):
        expr = self.parser.parse("not topic/a == 1")
        self.assertEqual(expr.keys, set(['topic/a']))
        self.assertFalse(expr({ 'topic/a' : 1 }))
        self.assertTrue(expr({ 'topic/a' : 2 }))
        self.assertTrue(expr.last_state)
        self.assertRaises(KeyError, expr, {})

    def test_and(self):
        expr = self.parser.parse("topic/a == 1 and topic/b == 2")
        self.assertEqual(expr.keys, set(['topic/a', 'topic/b']))
        self.assertTrue(expr({ 'topic/a' : 1, 'topic/b' : 2 }))
        self.assertFalse(expr({ 'topic/a' : 0, 'topic/b' : 2 }))
        self.assertFalse(expr({ 'topic/a' : 1, 'topic/b' : 0 }))
        self.assertFalse(expr.last_state)
        self.assertRaises(KeyError, expr, {})

    def test_or(self):
        expr = self.parser.parse("topic/a == 1 or topic/b == 2")
        self.assertEqual(expr.keys, set(['topic/a', 'topic/b']))
        self.assertTrue(expr({ 'topic/a' : 1, 'topic/b' : 1 }))
        self.assertTrue(expr({ 'topic/a' : 0, 'topic/b' : 2 }))
        self.assertFalse(expr({ 'topic/a' : 0, 'topic/b' : 0 }))
        self.assertFalse(expr.last_state)
        self.assertRaises(KeyError, expr, {})

    def test_precedence(self):
        expr = self.parser.parse("topic/b == 1 and topic/a == 2 or topic/a == 3 and topic/b == 4")
        self.assertEqual(expr.keys, set(['topic/a', 'topic/b']))
        self.assertTrue(expr({ 'topic/a' : 2, 'topic/b' : 1 }))
        self.assertTrue(expr({ 'topic/a' : 3, 'topic/b' : 4 }))
        self.assertFalse(expr({ 'topic/a' : 2, 'topic/b' : 3 }))
        self.assertFalse(expr.last_state)

    def test_parenthesis(self):
        expr = self.parser.parse("topic/a == 1 and (topic/b == 2 or topic/b == 3)")
        self.assertEqual(expr.keys, set(['topic/a', 'topic/b']))
        self.assertTrue(expr({ 'topic/a' : 1, 'topic/b' : 2 }))
        self.assertTrue(expr({ 'topic/a' : 1, 'topic/b' : 3 }))
        self.assertTrue(expr.last_state)

if __name__ == '__main__': unittest.main()



from datetime import datetime, timedelta
from localtime import *
from pprint import pprint
from tasks import Tasks
from unittest.mock import patch
import unittest, time


class TasksTest(unittest.TestCase):

    t1 = Tasks.Task(at(1), 't1', print)
    t2 = Tasks.Task(at(2), 't2', print)
    t3 = Tasks.Task(at(3), 't3', print)
    tasks = (t1, t2, t3) # sorted by time


    def setUp(self):
        self.sut = Tasks(0.01) # speed up for tests


    def tearDown(self):
        self.sut.stop()


    def test_add(self):
        self.sut.add(self.t1)
        self.assertIn(self.t1, self.sut)
        self.assertEqual(1, len(self.sut))
        
        self.sut.add(self.t1)
        self.assertIn(self.t1, self.sut)
        self.assertEqual(1, len(self.sut))
        
        self.sut.add(self.t2)
        self.assertIn(self.t2, self.sut)
        self.assertEqual(2, len(self.sut))

    
    def test_repr(self):
        self.sut.addAll(self.tasks)
        lines = repr(self.sut).split('\n')
        
        self.assertIn(str(self.t1.when), lines[1])
        self.assertIn(str(self.t2.when), lines[2])
        self.assertIn(str(self.t3.when), lines[3])
        
        
    @patch('tasks.now')
    def test_next(self, clock):
        self.sut.resolution = 0.1
        time.sleep(0.015)
        self.sut.addAll((self.t3, self.t1, self.t2))

        for expected in self.tasks:
            clock.return_value = expected.when
            task = self.sut.next()
            
            self.assertIsNotNone(task)
            self.assertEqual(expected, task)
            self.assertNotIn(task, self.sut)
            self.sut.resolution = 0.01
            

    def test_cancel(self):
        self.sut.addAll(self.tasks)
        self.sut.cancel('t2')
        self.assertNotIn(self.t2, self.sut)
        self.assertIn(self.t1, self.sut)
        self.assertIn(self.t3, self.sut)
        self.assertEqual(2, len(self.sut))


    @patch('tasks.now')
    def test_run(self, clock):            
        output = []
        t1 = Tasks.Task(at(0,1), 't1', lambda: output.append('t1'))
        t2 = Tasks.Task(at(0,2), 't2', lambda: output.append('t2'))
        t3 = Tasks.Task(at(0,3), 't3', lambda: output.append('t3'))        
        self.sut.addAll((t2, t1, t3))
        
        clock.return_value = t3.when
        time.sleep((len(self.sut) + 1) * self.sut.resolution)
        self.assertEqual(['t1', 't2', 't3'], output)
        

if __name__ == '__main__': unittest.main()
